/*********************/
/* Simple Deker test */
/*********************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdatomic.h>
#include <pthread.h>
#include "basic.h"

#define SZ (1024)
static int verbose = 0 ;

typedef struct {
  int sz ;
  volatile int x ;
  atomic_int want[2] ;
  atomic_int turn ;
  atomic_int start ;
} common_t ;

inline static void mbar(void) { __sync_synchronize(); }

inline static int other(int id) { return 1-id; }

inline static void lock(int id, common_t *q) {
  atomic_store(&q->want[id],1) ;
  while (atomic_load(&q->want[other(id)])) {
    if (atomic_load(&q->turn) !=id) {
      atomic_store(&q->want[id],0) ;
    }
    while (atomic_load(&q->turn) != id) ;
    atomic_store(&q->want[id],1) ;
  }
}

inline static void unlock(int id, common_t *q) {
  atomic_store_explicit(&q->turn,other(id),memory_order_release);
  atomic_store_explicit(&q->want[id],0,memory_order_release) ;
}

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *P(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int id = p->id ;
  int sz = q->sz ;

  while (!atomic_load_explicit(&q->start,memory_order_relaxed)) ;

  for (int k = sz ; k > 0 ; k--) {
    lock(id,q) ;
    delay(1) ;
    int tmp = q->x ;
    q->x = tmp+1 ;
    unlock(id,q) ;
  }
  return(NULL);
}


/************************/
/* Thread lauch and join */
/************************/

void run(int sz) {
  common_t c ;
  do {
    c.sz = sz ;
    atomic_init(&c.want[0],0) ;
    atomic_init(&c.want[1],0) ;
    atomic_init(&c.turn,0) ;
    c.x = 0 ;
    atomic_init(&c.start,0) ;

    pthread_t th[2] ; ctx_t a[2] ;
    for (int k = 0 ; k < 2 ; k++) {
      ctx_t *p = &a[k] ;
      p->id = k ; p->common = &c ;
      create_thread(&th[k],P,p) ;
    }
    atomic_store_explicit(&c.start,1,memory_order_relaxed) ;
    for (int k = 0 ; k < 2 ; k++) join_thread(&th[k]) ;
    printf("x=%i\n",c.x) ;
  } while  (c.x == 2 * c.sz) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}


int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int sz = SZ ;


  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    sz = atoi(argv[0]) ;
  }
  run(sz) ;
  return 0 ;
}

